from typing import Union
from functools import wraps
from datetime import datetime
import logging

from fastapi import HTTPException, Request, status
from authlib.oauth2.rfc6749 import OAuth2Token
from authlib.integrations.starlette_client import OAuth, OAuthError, StarletteOAuth2App

# from authlib.oauth1.auth import OAuthToken
# from authlib.oauth2.auth import OAuth2Token

from .models import User
from .database import db
from .settings import settings

logger = logging.getLogger(__name__)

OIDC_PROVIDERS = set([provider.id for provider in settings.oidc.providers])


def get_oidc_provider_or_none(request: Request) -> StarletteOAuth2App | None:
    """Return the oidc_provider from a request object, from the session.
    It can be used in Depends()"""
    if (oidc_provider_id := request.session.get("oidc_provider_id")) is None:
        return
    try:
        return getattr(authlib_oauth, str(oidc_provider_id))
    except AttributeError:
        return


def get_oidc_provider(request: Request) -> StarletteOAuth2App:
    if (oidc_provider := get_oidc_provider_or_none(request)) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No such provider")
    else:
        return oidc_provider


async def get_current_user(request: Request) -> User:
    """Get the current user from a request object.
    Also validates the token expiration time.
    ... TODO: complete about refresh token
    """
    if (user_sub := request.session.get("user_sub")) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED)
    if (token := await db.get_token(request.session["token"])) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Token unknown")
    user = await db.get_user(user_sub)
    ## Check if the token is expired
    if token.is_expired():
        oidc_provider = get_oidc_provider(request=request)
        ## Ask a new refresh token from the provider
        logger.info(f"Token expired for user {user.name}")
        try:
            userinfo = await oidc_provider.fetch_access_token(
                refresh_token=token.refresh_token
            )
        except OAuthError as err:
            logger.exception(err)
            # raise HTTPException(
            #    status.HTTP_401_UNAUTHORIZED, "Token expired, cannot refresh"
            # )

    return user


async def get_token(request: Request) -> OAuth2Token:
    """Return the token from a request object, from the session.
    It can be used in Depends()"""
    if (token := await db.get_token(request.session.get("token"))) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No token")
    return token


async def get_current_user_or_none(request: Request) -> User | None:
    """Return the user from a request object, from the session.
    It can be used in Depends()"""
    try:
        return await get_current_user(request)
    except HTTPException:
        return None


def hasrole(required_roles: Union[str, list[str]] = []):
    """Decorator for RBAC permissions"""
    required_roles_set: set[str]
    if isinstance(required_roles, str):
        required_roles_set = set([required_roles])
    else:
        required_roles_set = set(required_roles)

    def decorator(func):
        @wraps(func)
        async def wrapper(request=None, *args, **kwargs):
            if request is None:
                raise HTTPException(
                    500,
                    "Functions decorated with hasrole must have a request:Request argument",
                )
            user: User = await get_current_user(request)
            if not any(required_roles_set.intersection(user.roles_as_set)):
                raise HTTPException(status.HTTP_401_UNAUTHORIZED)
            return await func(request, *args, **kwargs)

        return wrapper

    return decorator


def get_token_info(token: dict) -> dict:
    token_info = dict()
    for key in token:
        if key != "userinfo":
            token_info[key] = token[key]
    return token_info


def fetch_token(name, request):
    breakpoint()
    ...
    # if name in OIDC_PROVIDERS:
    #    model = OAuth2Token
    # else:
    #    model = OAuthToken

    # token = model.find(name=name, user=request.user)
    # return token.to_token()


def update_token(*args, **kwargs):
    breakpoint()
    ...


authlib_oauth = OAuth(cache=None, fetch_token=fetch_token, update_token=update_token)
